"""
:author: Gatsby Lee
:since: 2019-08-09
"""
import hashlib
import os
import gzip
import shutil

LOCAL_STORAGE_ROOT = '/tmp/local_storage'


class LocalStorage(object):

    def __init__(self, root_dir=LOCAL_STORAGE_ROOT):
        self._root_dir = root_dir
        if not os.path.exists(self._root_dir):
            os.mkdir(self._root_dir)

    def _get_full_path(self, bucket_name, object_key):
        object_key_hash = hashlib.md5(object_key.encode("utf-8")).hexdigest()
        bucket_directory = os.path.join(self._root_dir, bucket_name)
        if not os.path.exists(bucket_directory):
            self.create_bucket(bucket_name)
        return os.path.join(self._root_dir, bucket_name, object_key_hash)

    def create_bucket(self, bucket_name):
        """
        Create bucket
        """
        bucket_path = os.path.join(self._root_dir, bucket_name)
        os.mkdir(bucket_path)

    def list_bucket_names(self):
        """Return list of bucket names

        Returns:
            bucket_names(list)
        """
        bucket_names = []
        for x in os.listdir(self._root_dir):
            if os.path.isdir(os.path.join(self._root_dir, x)):
                bucket_names += x,
        return bucket_names

    def upload_file(self, bucket_name, object_key, source_file_name,
                    content_type=None, content_encoding=None):
        """
        Uploads a local file to a bucket
        """
        full_path = self._get_full_path(bucket_name, object_key)
        if content_encoding == 'gzip':
            with gzip.open(full_path, 'wb') as fw:
                with open(source_file_name, 'rb') as fr:
                    fw.write(fr.read())
        else:
            shutil.copy(source_file_name, full_path)

    def upload(self, bucket_name, object_key, buffer,
               content_type='', content_encoding=''):
        """Upload content to a bucket

        Args:
            bucket_name(str):  Bucket name to use
            buffer(bytes): Content to upload
            object_key(str): Object key stored in bucket
        Kwargs:
            content_type(str): Type of Content
            content_encoding(str): Encoding used on content for uploading
        Returns:
            None
        """
        assert isinstance(buffer, bytes)

        full_path = self._get_full_path(bucket_name, object_key)
        if content_encoding == 'gzip':
            with gzip.open(full_path, 'wb') as fw:
                fw.write(buffer)
        else:
            with open(full_path, 'wb') as fw:
                fw.write(buffer)

    def is_exists(self, bucket_name, object_key):
        """Check if an object exists in bucket

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object key stored in bucket

        Returns:
            True if exists.
        """
        full_path = self._get_full_path(bucket_name, object_key)
        return os.path.exists(full_path)

    def rename(self, bucket_name, object_key, new_object_key):
        """Renames an object

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object Key to rename
            new_object_key (str): Object Key to be renamed to
        Returns:
            None
        """
        assert(object_key != new_object_key), \
            "object_key can't be same to new_object_key"

        shutil.move(
            self._get_full_path(bucket_name, object_key),
            self._get_full_path(bucket_name, new_object_key),
        )

    def download_gzipped_to_file(self, bucket_name, object_key, destination_file_name,
                                 do_gunzip=False):
        """Download an object to local

         Args:
             bucket_name(str):  Bucket name to use
             object_key(str): Object Key to rename
             destination_file_name(str): Local file path
        Kwargs:
            do_gunzip(bool): True to gunzip(default: False)
        Returns:
            None
         """
        src_full_path = self._get_full_path(bucket_name, object_key)
        dest_full_path = self._get_full_path(bucket_name, object_key)
        if do_gunzip:
            with gzip.open(src_full_path, 'rb') as fr:
                with open(dest_full_path, 'wb') as fw:
                    fw.write(fr.read())
        else:
            shutil.copy(src_full_path, dest_full_path)

    def download_gzipped(self, bucket_name, object_key, do_gunzip=False):
        """Download an gzipped object content to memory

        Args:
            bucket_name(str):  Bucket name to use
            object_key(str): Object Key to rename
        Kwargs:
            do_gunzip(bool): True to gunzip(default: False)
        Returns:
            bytes. Content stored in the object
        """
        full_path = self._get_full_path(bucket_name, object_key)
        if do_gunzip:
            with gzip.open(full_path, 'rb') as fr:
                return fr.read()
        else:
            with open(full_path, 'rb') as fr:
                return fr.read()

    def delete(self, bucket_name, object_key):
        """Delete an object from bucket

        Args:
            bucket_name(str):  Bucket name to use
            object_key(str): Object Key to rename
        Returns:
            None
        @note: No exception raises although object doesn't exist.
        """
        full_path = self._get_full_path(bucket_name, object_key)
        os.remove(full_path)



from cloud_storage.__about__ import __version__
from cloud_storage.enums import CloudStorageType
from cloud_storage.excepts import UnsupportedStorage
from cloud_storage.local_storage import LocalStorage
from cloud_storage.gcs_storage import GoogleCloudStorage
from cloud_storage.s3_storage_boto3 import S3CloudStorageBoto3

STORAGE_CLIENT_MAPPING = {
    CloudStorageType.GCS: GoogleCloudStorage,
    CloudStorageType.LOCAL: LocalStorage,
    CloudStorageType.S3: S3CloudStorageBoto3,
}


def create_storage_client(name):
    """
    Return new CloudStorage client

    Args:
        name
    """
    try:
        client_class = STORAGE_CLIENT_MAPPING[name]
        return client_class()
    except KeyError:
        raise UnsupportedStorage('%s is not supported.' % name)


def int_or_str(value):
    try:
        return int(value)
    except ValueError:
        return value


VERSION = tuple(map(int_or_str, __version__.split('.')))

__all__ = (
    '__version__',
    'VERSION',
    'create_storage_client',
    'LocalStorage',
    'GoogleCloudStorage',
    'S3CloudStorageBoto3',
)



"""
@author Henley Kuang
@since 2019-06-13

@note: classes are in alphabetical order
"""


class UnsupportedStorage(Exception):
    pass


class CloudStorageBadRequestException(Exception):
    pass


class CloudStorageInvalidArgumentTypeException(Exception):
    pass


class CloudStorageNotFoundException (Exception):
    pass


class CloudStorageServerErrorException(Exception):
    pass


class CloudStorageUnknownErrorException(Exception):
    pass


__all__ = (
    'CloudStorageBadRequestException',
    'CloudStorageInvalidArgumentTypeException',
    'CloudStorageNotFoundException',
    'CloudStorageServerErrorException',
    'CloudStorageUnknownErrorException',
)



"""
:author: Henley Kuang
:since: 06/12/2019
"""
import logging
import os

import google.api_core.exceptions
import google.cloud.storage

from google import resumable_media

from cloud_storage.excepts import (
    CloudStorageBadRequestException,
    CloudStorageInvalidArgumentTypeException,
    CloudStorageNotFoundException,
    CloudStorageServerErrorException,
    CloudStorageUnknownErrorException,
)


"""
@ref:
https://2.python-requests.org//en/master/user/quickstart/#binary-response-content
https://github.com/googleapis/google-resumable-media-python/blob/master/google/resumable_media/requests/__init__.py
https://github.com/googleapis/google-resumable-media-python/blob/master/google/resumable_media/requests/download.py#L120
https://github.com/googleapis/google-resumable-media-python/blob/master/google/resumable_media/requests/download.py#L139
https://github.com/googleapis/google-cloud-python/blob/master/storage/google/cloud/storage/blob.py
"""

LOGGER = logging.getLogger(__name__)


def gcs_api_exception_handler(f):
    def decorate(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except google.api_core.exceptions.BadRequest as e:
            raise CloudStorageBadRequestException(str(e))
        except (
            google.api_core.exceptions.InternalServerError,
            google.api_core.exceptions.ServerError,
            google.api_core.exceptions.ServiceUnavailable,
        ) as e:
            raise CloudStorageServerErrorException(str(e))
        except google.api_core.exceptions.NotFound as e:
            raise CloudStorageNotFoundException(str(e))
        except AssertionError as e:
            raise CloudStorageInvalidArgumentTypeException(str(e)) from e
        except Exception as e:
            raise CloudStorageUnknownErrorException(str(e)) from e

    return decorate


class GoogleCloudStorage(object):
    def __init__(self):
        self.storage_client = google.cloud.storage.Client()
        self._buckets = {}

    def _get_bucket(self, bucket_name):
        try:
            return self._buckets[bucket_name]
        except KeyError:
            self._buckets[bucket_name] = self.storage_client.get_bucket(bucket_name)
            return self._buckets[bucket_name]

    def list_bucket_names(self):
        """Get the list of buckets in GCS
        Returns:
            list. List of strings of bucket names

        Example.

        >>> GoogleCloudStorage().list_bucket_names()
        ["bucket1", "bucket2"]
        """
        return list(self.storage_client.list_buckets())

    @gcs_api_exception_handler
    def upload_file(
        self,
        bucket_name,
        object_key,
        source_file_name,
        content_type=None,
        content_encoding=None,
    ):
        """Upload a file to a bucket

        Args:
            bucket_name (str):  Bucket name to use
            source_file_name (str): Local file path
            object_key (str): Object key stored in bucket
        Kwargs:
            content_type (str): Type of Content
            content_encoding(str): Encoding used on content for uploading
        Returns:
            None
        """
        bucket = self._get_bucket(bucket_name)
        blob = bucket.blob(object_key)
        if content_encoding is not None:
            blob.content_encoding = content_encoding

        blob.upload_from_filename(source_file_name, content_type=content_type)

    @gcs_api_exception_handler
    def upload(
        self, bucket_name, object_key, buffer, content_type=None, content_encoding=None
    ):
        """Upload content to a bucket

        Args:
            bucket_name (str):  Bucket name to use
            buffer (bytes): Content to upload
            object_key (str): Object key stored in bucket
        Kwargs:
            content_type (str): Type of Content
            content_encoding(str): Encoding used on content for uploading
        Returns:
            None
        """
        assert isinstance(buffer, bytes)
        bucket = self._get_bucket(bucket_name)
        blob = bucket.blob(object_key)

        if content_encoding is not None:
            blob.content_encoding = content_encoding

        blob.upload_from_string(buffer, content_type=content_type)

    @gcs_api_exception_handler
    def is_exists(self, bucket_name, object_key):
        """Check if an object exists in bucket

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object key stored in bucket

        Returns:
            boolean. Values::
                True -- Exists in GCS
                False -- Does not exist in GCS

        Example.

        >>> GoogleCloudStorage().is_exists("my_bucket", "my_object_key")
        True
        """
        bucket = self._get_bucket(bucket_name)
        blob = bucket.blob(object_key)
        return blob.exists()

    @gcs_api_exception_handler
    def rename(self, bucket_name, object_key, new_object_key):
        """Renames an object

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object Key to rename
            new_object_key (str): Object Key to be renamed to
        Returns:
            None
        """
        assert (
            object_key != new_object_key
        ), "object_key can't be same to new_object_key"

        bucket = self._get_bucket(bucket_name)
        blob = bucket.blob(object_key)
        # Since this is not atomic operation, deleting current key might fail.
        bucket.rename_blob(blob, new_object_key)

    @gcs_api_exception_handler
    def download_gzipped_to_file(
        self, bucket_name, object_key, destination_file_name, do_gunzip=False
    ):
        """Download an object to local

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object Key to rename
            destination_file_name (str): Local file path
        Kwargs:
            do_gunzip(bool): True to gunzip(default: False)
        Returns:
            None
        """
        bucket = self._get_bucket(bucket_name)
        blob = bucket.blob(object_key)

        # blob.download_to_filename(destination_file_name)
        # https://googleapis.github.io/google-cloud-python/latest/_modules/google/cloud/storage/blob.html#Blob.download_to_file
        raw_download = not do_gunzip
        try:
            with open(destination_file_name, "wb") as file_obj:
                blob.download_to_file(file_obj, raw_download=raw_download)
        except resumable_media.DataCorruption:
            # Delete the corrupt downloaded file.
            os.remove(destination_file_name)
            raise

    @gcs_api_exception_handler
    def download_gzipped(self, bucket_name, object_key, do_gunzip=False):
        """Download an object to memory

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object Key to rename
        Kwargs:
            do_gunzip(bool): True to gunzip(default: False)
        Returns:
            bytes. Content stored in the object

        Example.
        >>> GoogleCloudStorage().download_gzipped("my_bucket", "my_object_key")
        b'This is content in your object key'
        """
        bucket = self._get_bucket(bucket_name)
        blob = bucket.blob(object_key)

        # Google uses python-requests.
        # And, The gzip and deflate transfer-encodings are automatically decoded.
        # ref: https://github.com/googleapis/google-resumable-media-python/blob/master/google/resumable_media/requests/download.py#L120
        if do_gunzip:
            return blob.download_as_string()

        return blob.download_as_string(raw_download=True)

    @gcs_api_exception_handler
    def delete(self, bucket_name, object_key):
        """Delete an object from bucket.

        Exception won't be raised although object_key doesn't exist.

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object Key to rename
        Returns:
            None
        """
        bucket = self._get_bucket(bucket_name)
        blob = bucket.blob(object_key)
        try:
            blob.delete()
        except google.api_core.exceptions.NotFound:
            # slience if object_key doesn't exists
            pass



"""
:author: Gatsby Lee
:since: 06/21/2019
"""
import gzip
import io
import logging

import boto3
import botocore

from http import HTTPStatus

from cloud_storage.excepts import (
    CloudStorageInvalidArgumentTypeException,
    CloudStorageNotFoundException,
    CloudStorageUnknownErrorException,
)

LOGGER = logging.getLogger(__name__)

"""
https://stackoverflow.com/questions/42809096/difference-in-boto3-between-resource-client-and-session
https://boto3.amazonaws.com/v1/documentation/api/latest/guide/clients.html
"""


def s3_boto3_api_exception_handler(f):
    def decorate(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except botocore.exceptions.ClientError as e:
            if e.response['ResponseMetadata']['HTTPStatusCode'] == HTTPStatus.NOT_FOUND:
                raise CloudStorageNotFoundException(str(e)) from None
            # not handled, bring up.
            raise e
        except AssertionError as e:
            raise CloudStorageInvalidArgumentTypeException(str(e)) from e
        except Exception as e:
            raise CloudStorageUnknownErrorException(str(e)) from e
    return decorate


class S3CloudStorageBoto3(object):

    def __init__(self):
        self.storage_client = boto3.client('s3')

    def list_bucket_names(self):
        """Return list of bucket names

        Returns:
            bucket_names(list)
        """
        api_response = self.storage_client.list_buckets()
        bucket_raw_info_list = api_response['Buckets']
        bucket_names = []
        for bucket_raw_info in bucket_raw_info_list:
            bucket_names.append(bucket_raw_info['Name'])
        return bucket_names

    def upload_file(self, bucket_name, object_key, source_file_name,
                    content_type=None, content_encoding=None):
        """
        Uploads a local file to a bucket
        """
        # Allowed values for ExtraArgs
        # ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.S3Transfer.ALLOWED_UPLOAD_ARGS
        self.storage_client.upload_file(
            Bucket=bucket_name, Key=object_key, Filename=source_file_name,
            ExtraArgs={'ContentType': content_type,
                       'ContentEncoding': content_encoding}
        )

    def upload(self, bucket_name, object_key, buffer,
               content_type='', content_encoding=''):
        """Upload content to a bucket

        Args:
            bucket_name(str):  Bucket name to use
            buffer(bytes): Content to upload
            object_key(str): Object key stored in bucket
        Kwargs:
            content_type(str): Type of Content
            content_encoding(str): Encoding used on content for uploading
        Returns:
            None
        """
        assert isinstance(buffer, bytes)
        self.storage_client.put_object(
            Bucket=bucket_name, Key=object_key, Body=buffer,
            ContentType=content_type, ContentEncoding=content_encoding
        )

    def is_exists(self, bucket_name, object_key):
        """Check if an object exists in bucket

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object key stored in bucket

        Returns:
            True if exists.
        """
        try:
            self.storage_client.head_object(Bucket=bucket_name, Key=object_key)
            return True
        except botocore.exceptions.ClientError as e:
            if e.response['ResponseMetadata']['HTTPStatusCode'] == HTTPStatus.NOT_FOUND:
                return False
            # not handled, bring up.
            raise e

    @s3_boto3_api_exception_handler
    def rename(self, bucket_name, object_key, new_object_key):
        """Renames an object

        Args:
            bucket_name (str):  Bucket name to use
            object_key (str): Object Key to rename
            new_object_key (str): Object Key to be renamed to
        Returns:
            None
        """
        assert(object_key != new_object_key), \
            "object_key can't be same to new_object_key"

        # Since this is not atomic operation, deleting current key might fail.
        self.storage_client.copy_object(
            Bucket=bucket_name, Key=new_object_key,
            CopySource={
                'Bucket': bucket_name, 'Key': object_key},
        )
        self.delete(bucket_name, object_key)

    @s3_boto3_api_exception_handler
    def download_gzipped_to_file(self, bucket_name, object_key, destination_file_name,
                                 do_gunzip=False):
        """Download an object to local

         Args:
             bucket_name(str):  Bucket name to use
             object_key(str): Object Key to rename
             destination_file_name(str): Local file path
        Kwargs:
            do_gunzip(bool): True to gunzip(default: False)
        Returns:
            None
         """
        if do_gunzip:
            object_content = self.download_gzipped(
                bucket_name, object_key, do_gunzip)
            with open(destination_file_name, 'wb') as f:
                f.write(object_content)
        else:
            self.storage_client.download_file(
                Bucket=bucket_name, Key=object_key, Filename=destination_file_name)

    @s3_boto3_api_exception_handler
    def download_gzipped(self, bucket_name, object_key, do_gunzip=False):
        """Download an gzipped object content to memory

        Args:
            bucket_name(str):  Bucket name to use
            object_key(str): Object Key to rename
        Kwargs:
            do_gunzip(bool): True to gunzip(default: False)
        Returns:
            bytes. Content stored in the object
        """
        if do_gunzip:
            with io.BytesIO() as bytes_buffer:
                with gzip.GzipFile(fileobj=bytes_buffer, mode='rb') as rgzip_buffer:
                    self.storage_client.download_fileobj(
                        Bucket=bucket_name, Key=object_key, Fileobj=bytes_buffer)
                    bytes_buffer.seek(0)
                    return rgzip_buffer.read()

        with io.BytesIO() as bytes_buffer:
            self.storage_client.download_fileobj(
                Bucket=bucket_name, Key=object_key, Fileobj=bytes_buffer)
            return bytes_buffer.getvalue()

    @s3_boto3_api_exception_handler
    def delete(self, bucket_name, object_key):
        """Delete an object from bucket

        Args:
            bucket_name(str):  Bucket name to use
            object_key(str): Object Key to rename
        Returns:
            None
        @note: No exception raises although object doesn't exist.
        """
        self.storage_client.delete_object(
            Bucket=bucket_name,
            Key=object_key,
        )



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

version = {}
with open(path.join(here, "cloud_storage/__about__.py"), encoding="utf-8") as f:
    exec(f.read(), version)

requires = [
    "boto3",
    "google-cloud-storage >= 1.24.0",
    "six >= 1.15.0",
]

test_requires = [
    "pytest",
    "pytest-mock",
    "moto",
]

dev_requires = test_requires + [
    "wheel",
    "bpython",
    "pylint",
    "black",
]

setup(
    name="cloud-storage",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # http://packaging.python.org/en/latest/tutorial.html#version
    version=version["__version__"],
    description="Interface of S3 and Google Storgae",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # The project's main homepage.
    url="https://github.com/Gatsby-Lee/cloud-storage",
    # Author details
    author="Henley Kuang, Gatsby Lee",
    # Choose your license
    license="GPL-3.0",
    # See https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Utilities",
    ],
    # What does your project relate to?
    keywords="cloud storage gcs s3",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "docs", "tests*", "playground*"]),
    # List run-time dependencies here.  These will be installed by pip when your
    # project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/technical.html#install-requires-vs-requirements-files
    python_requires=">=3.6",
    install_requires=requires,
    extras_require={
        "dev": dev_requires,
        "test": test_requires,
    },
)
